# $Id: emoncms.py 1143 2014-11-25 19:54:30Z mwall $
# Copyright 2013-2014 Matthew Wall

"""
Emoncms is a powerful open-source web-app for processing, logging and
visualising energy, temperature and other environmental data.

http://emoncms.org

This is a weewx extension that uploads data to an EmonCMS server.

Minimal Configuration

A read/write token is required.  The default configuration will upload to
emoncms.org.  All weewx variables will be uploaded using weewx names and
default units and formatting.

[StdRESTful]
    [[EmonCMS]]
        token = TOKEN

Customized Configuration

When an input map is specified, only variables in that map will be uploaded.
The 'units' parameter can be used to specify which units should be used for
the input, independent of the local weewx units.

[StdRESTful]
    [[EmonCMS]]
        token = TOKEN
        prefix = weather
        server_url = http://192.168.0.1/emoncms/input/post.json
        [[[inputs]]]
            [[[[barometer]]]]
                units = inHg
                name = barometer_inHg
                format = %.3f
            [[[[outTemp]]]]
                units = degree_F
                name = outTemp_F
                format = %.1f
            [[[[outHumidity]]]]
                name = outHumidity
                format = %03.0f
            [[[[windSpeed]]]]
                units = mph
                name = windSpeed_mph
                format = %.2f
            [[[[windDir]]]]
                format = %03.0f
"""

# FIXME: do pattern matching for many similarly-named channels

import Queue
import sys
import syslog
import urllib
import urllib2

import weewx
import weewx.restx
import weewx.units
from weeutil.weeutil import to_bool, accumulateLeaves

VERSION = "0.6rc2"

if weewx.__version__ < "3":
    raise weewx.UnsupportedFeature("weewx 3 is required, found %s" %
                                   weewx.__version__)

def logmsg(level, msg):
    syslog.syslog(level, 'restx: EmonCMS: %s' % msg)

def logdbg(msg):
    logmsg(syslog.LOG_DEBUG, msg)

def loginf(msg):
    logmsg(syslog.LOG_INFO, msg)

def logerr(msg):
    logmsg(syslog.LOG_ERR, msg)

def _obfuscate(s):
    return ('X'*(len(s)-4) + s[-4:])

def _compat(d, old_label, new_label):
    if d.has_key(old_label) and not d.has_key(new_label):
        d.setdefault(new_label, d[old_label])
        d.pop(old_label)

class EmonCMS(weewx.restx.StdRESTbase):
    def __init__(self, engine, config_dict):
        """This service recognizes standard restful options plus the following:

        Required parameters:

        token: unique token for read-write access to emoncms

        Optional parameters:

        node: integer value for emoncms node
        Default is None
        
        prefix: label that will be pre-pended to each variable
        Default is None

        server_url: URL of the server
        Default is emoncms.org

        inputs: dictionary of weewx observation names with optional upload
        name, format, and units
        Default is None
        """
        super(EmonCMS, self).__init__(engine, config_dict)        
        loginf("service version is %s" % VERSION)
        try:
            site_dict = config_dict['StdRESTful']['EmonCMS']
            site_dict = accumulateLeaves(site_dict, max_level=1)
            site_dict['token']
        except KeyError, e:
            logerr("Data will not be posted: Missing option %s" % e)
            return

        try:
            if site_dict.has_key('node'):
                int(site_dict['node'])
        except (TypeError, ValueError), e:
            logerr("Data will not be posted: Node must be integer: %s" % e)
            return

        if config_dict['StdRESTful']['EmonCMS'].has_key('inputs'):
            site_dict['inputs'] = dict(config_dict['StdRESTful']['EmonCMS']['inputs'])

        site_dict['manager_dict'] = weewx.manager.get_manager_dict(
            config_dict['DataBindings'], config_dict['Databases'], 'wx_binding')

        # for backward compatibility: 'url' is now 'server_url'
        _compat(site_dict, 'url', 'server_url')
        # for backward compatibility: 'station' is now 'prefix'
        _compat(site_dict, 'station', 'prefix')

        self.archive_queue = Queue.Queue()
        self.archive_thread = EmonCMSThread(self.archive_queue, **site_dict)
        self.archive_thread.start()
        self.bind(weewx.NEW_ARCHIVE_RECORD, self.new_archive_record)
        if 'prefix' in site_dict:
            loginf("prefix is %s" % site_dict['prefix'])
        if 'node' in site_dict:
            loginf("node is %s" % site_dict['node'])
        if 'unit_system' in site_dict:
            loginf("desired unit system is %s" % site_dict['unit_system'])
        if 'inputs' in site_dict:
            loginf("using specified input map")
        loginf("Data will be uploaded with token=%s" %
               _obfuscate(site_dict['token']))

    def new_archive_record(self, event):
        self.archive_queue.put(event.record)

class EmonCMSThread(weewx.restx.RESTThread):

    _DEFAULT_SERVER_URL = 'http://emoncms.org/input/post.json'

    def __init__(self, queue, token, manager_dict,
                 prefix=None, node=None, unit_system=None,
                 server_url=_DEFAULT_SERVER_URL, inputs=None,
                 skip_upload=False,
                 post_interval=None, max_backlog=sys.maxint, stale=None,
                 log_success=True, log_failure=True,
                 timeout=60, max_tries=3, retry_wait=5):
        super(EmonCMSThread, self).__init__(queue,
                                            protocol_name='EmonCMS',
                                            manager_dict=manager_dict,
                                            post_interval=post_interval,
                                            max_backlog=max_backlog,
                                            stale=stale,
                                            log_success=log_success,
                                            log_failure=log_failure,
                                            max_tries=max_tries,
                                            timeout=timeout,
                                            retry_wait=retry_wait)
        self.token = token
        self.prefix = prefix
        self.node = node
        self.server_url = server_url
        self.inputs = inputs
        self.skip_upload = to_bool(skip_upload)
        self.units = weewx.units.unit_constants[unit_system] if unit_system is not None else None

    def process_record(self, record, dbm):
        r = self.get_record(record, dbm)
        url = self.get_url(r)
        if self.skip_upload:
            logdbg("skipping upload")
            return
        req = urllib2.Request(url)
        req.add_header("User-Agent", "weewx/%s" % weewx.__version__)
        self.post_with_retries(req)

    def check_response(self, response):
        txt = response.read()
        if txt != 'ok' :
            raise weewx.restx.FailedPost("Server returned '%s'" % txt)

    def get_url(self, record):
        # if there is a prefix, prepend it to every variable name
        prefix = ''
        if self.prefix is not None:
            prefix = '%s_' % urllib.quote_plus(self.prefix)

        # convert to desired unit system if necessary
        if self.units is not None and record['usUnits'] != self.units:
            converter = weewx.units.StdUnitConverters[self.units]
            record = converter.convertDict(record)

        # if there is an input map, use it to map the weewx variables into
        # emoncms inputs.  otherwise, use the raw weewx names and default
        # formatting and units.
        data = []
        if self.inputs is not None:
            for k in self.inputs:
                try:
                    v = float(record.get(k))
                    name = self.inputs[k].get('name', k)
                    fmt = self.inputs[k].get('format', '%s')
                    to_units = self.inputs[k].get('units')
                    if to_units is not None:
                        (from_unit, from_group) = weewx.units.getStandardUnitType(record['usUnits'], k)
                        from_t = (v, from_unit, from_group)
                        v = weewx.units.convert(from_t, to_units)[0]
                    s = fmt % v
                    data.append('%s%s:%s' % (prefix, name, s))
                except (TypeError, ValueError):
                    pass
        else:
            for k in record:
                try:
                    v = float(record[k])
                    data.append('%s%s:%s' % (prefix, k, v))
                except (TypeError, ValueError):
                    pass
        # assemble the complete url from all the inputs and values
        parts = []
        parts.append(self.server_url)
        parts.append('?apikey=%s' % self.token)
        parts.append('&time=%s' % record['dateTime'])
        if self.node is not None:
            parts.append('&node=%d' % self.node)
        parts.append('&json={%s}' % ','.join(data))
        url = ''.join(parts)
        logdbg('url: %s' % url)
        return url
